@ECHO OFF
REM to check error codes in use: grep -o "el=[0-9]\+" [file]|sort -t= -k2 -n
SETLOCAL ENABLEEXTENSIONS ENABLEDELAYEDEXPANSION
IF /I "%USERPROFILE%" NEQ "C:\Users\%USERNAME%" (SET MYDOCS=My Documents) ELSE SET MYDOCS=Documents
SET findprog=%SYSTEMROOT%\system32\find.exe
SET debug=
SET quiet=n
SET MYAPPDATA=%APPDATA%\TimeDicer
SET ver=8.14
SET credit=Dominic Raferd
SET contact=dominic@timedicer.co.uk
SET homepage=https://www.timedicer.co.uk
SET rdiffbackup_on_client=rdiff-backup.exe
SET rdiffbackup_on_server=rdiff-backup
SET myid=%random%%random%
SET datetime=%DATE% %TIME:~0,5%
REM set our own errorlevel indicator to 0
SET el=0

REM change current path to the location of %this% - could use PUSHD instead?
SET execpath=%~dp0
SET execpath=%execpath:~0,-1%
SET startpath=%CD%
IF /I NOT DEFINED execpath SET execpath=%CD%
IF /I "%execpath%" NEQ "%CD%" CD /D %execpath%
IF ERRORLEVEL 1 SET el=116&&GOTO :endd
SET runfrom=%CD%
SET this=%~n0
CALL :lower_case "%this%"
IF ERRORLEVEL 1 SET el=118&&GOTO :endd
SET thislc=%lcase%

SET log1=%TEMP%\%thislc%-%myid%-log1.txt
SET log2=%TEMP%\%thislc%-%myid%-log2.txt
REM set default logfile
SET log=%MYAPPDATA%\%thislc%-log.html
REM set default action
SET action=Backing up


REM ---- Establish variables and configuration files ----
REM set default user value to %USERNAME%-%USERDOMAIN% but in lowercase with spaces stripped
CALL :lower_case "%USERNAME%-%USERDOMAIN%"
SET user=%lcase%
REM set default key value
SET key="%MYAPPDATA%\privatekey.ppk"
IF NOT EXIST %key% SET key="%USERPROFILE%\privatekey.ppk"
REM set default usage of Volume Shadow Services
SET vss=y
REM set default port for server
SET port=22
REM set default action for existing shadow copies
SET deleteshadows=n
REM set default skipping of backup if source has not changed since last time
SET skip_on_no_change=y
REM set default rdiff-backup options
SET options=--no-acls -v5
IF ERRORLEVEL 1 SET el=131&&GOTO :endd

REM ---- Process command line options ----
REM g for debug messages
CALL :param_check g %*
IF ERRORLEVEL 1 SET el=119&&GOTO :endd
IF %param% GTR 0 (
	SET debug=y
	ECHO Debug mode
	ECHO CD: "%CD%"
	ECHO runfrom: "%runfrom%"
)

REM ---- Check for presence of required external files ----
REM we use file_check for program files that must be in the current path. Although
REM not strictly necessary that all these files be here provided
REM they can be run from the command line, it is tidier to keep them in the
REM same directory as this command script.
IF ERRORLEVEL 1 SET el=,4&&GOTO :endd
CALL :file_check %rdiffbackup_on_client% https://github.com/rdiff-backup/rdiff-backup %el%
IF ERRORLEVEL 1 SET el=5&&GOTO :endd
CALL :file_check Microsoft.VC90.CRT.manifest https://www.microsoft.com/en-us/download/details.aspx?id=26368 %el%
IF ERRORLEVEL 1 SET el=5&&GOTO :endd
CALL :file_check plink.exe http://www.chiark.greenend.org.uk/~sgtatham/putty/download.html %el%
IF ERRORLEVEL 1 SET el=5&&GOTO :endd
IF %el% GEQ 1 GOTO :endd
CALL :file_check ssed.exe http://sed.sourceforge.net/grabbag/ssed/sed-3.62.zip %el%
IF ERRORLEVEL 1 SET el=5&&GOTO :endd
IF DEFINED debug ECHO All required external files were found

REM f to set configuration file
CALL :param_check f %*
IF ERRORLEVEL 1 SET el=127&&GOTO :endd
IF %param% GTR 0 (
	SET configfile=%nextparam%
) ELSE (
	SET configfile="%MYAPPDATA%\%thislc%.txt"
)
IF DEFINED debug ECHO configfile identified as: %configfile%
REM configfl is configfile dequoted
SET configfl=%configfile%
CALL :dequote configfl
IF DEFINED debug ECHO configfl set to: "%configfl%"
REM set configfile to shortname format
CALL:shortformdequote configfile
IF DEFINED debug ECHO configfile reset to: "%configfile%"
IF EXIST %configfile% (
	IF DEFINED debug ECHO about to create %TEMP%\%thislc%-%myid%-setup.bat using "%runfrom%\ssed.exe" on configfile
	"%runfrom%\ssed.exe" -n "/^\s*SET /I{s/\s*#.*//;/^^$/^!p}" "!configfile!">%TEMP%\%thislc%-%myid%-setup.bat
	IF DEFINED debug ECHO about to create %thislc%-%myid%-datasets.txt
	"%runfrom%\ssed.exe" -n "/^^\s*SET /I^!{s/\s*#.*//;/^^$/^!p}" "!configfile!">%TEMP%\%thislc%-%myid%-dataset1.txt
	IF DEFINED debug ECHO about to run %TEMP%\%thislc%-%myid%-setup.bat
	CALL %TEMP%\%thislc%-%myid%-setup.bat
) ELSE (
	SET el=10&&GOTO :endd
)
IF DEFINED debug ECHO here0

REM v for version
CALL :param_check v %*
IF ERRORLEVEL 1 SET el=120&&GOTO :endd
IF DEFINED debug echo here2
IF %param% GTR 0 ECHO %ver%&&GOTO :eof

REM h or ? for help
CALL :param_check h %*
IF ERRORLEVEL 1 SET el=121&&GOTO :endd
IF %param% GTR 0 GOTO :help
CALL :param_check ? %*
IF ERRORLEVEL 1 SET el=122&&GOTO :endd
IF %param% GTR 0 GOTO :help

IF %quiet% == n ECHO.&ECHO %this% v%ver%&ECHO.
ECHO ^<h2^>^<a name="%datetime%"^>^</a^>Session %datetime%^</h2^>>%log1%

REM x for exit at end, not pause
CALL :param_check x %*
IF ERRORLEVEL 1 SET el=146&&GOTO :endd
IF %param% GTR 0 (SET exitnow=y) ELSE SET exitnow=n

REM d for force deletion of existing shadow sets
CALL :param_check d %*
IF ERRORLEVEL 1 SET el=123&&GOTO :endd
IF %param% GTR 0 (SET forcedeleteshadows=y) ELSE SET forcedeleteshadows=n

REM q for quiet
CALL :param_check q %*
IF ERRORLEVEL 1 SET el=124&&GOTO :endd
IF %param% GTR 0 (SET quiet=y) ELSE SET quiet=n

REM s to set server
IF DEFINED debug ECHO here3
CALL :param_check s %*
IF ERRORLEVEL 1 SET el=148&&GOTO :endd
IF %param% GTR 0 (
	REM remove any quotes
	SET server=%nextparam%
	CALL :dequote server
	IF DEFINED debug ECHO server identified as "!server!"
)

REM u to set user
IF DEFINED debug ECHO here4
CALL :param_check u %*
IF ERRORLEVEL 1 SET el=149&&GOTO :endd
IF %param% GTR 0 (
	REM remove any quotes
	SET user=%nextparam%
	CALL :dequote user
	IF DEFINED debug ECHO user identified as "!user!"
)

REM k to set key file
IF DEFINED debug ECHO here5
CALL :param_check k %*
IF ERRORLEVEL 1 SET el=150&&GOTO :endd
IF DEFINED debug ECHO here6
IF %param% GTR 0 (
	SET key=%nextparam%
	IF DEFINED debug ECHO key identified as "!key!"
)

REM o to set options
CALL :param_check o %*
IF ERRORLEVEL 1 SET el=155&&GOTO :endd
IF DEFINED debug ECHO here7
IF %param% GTR 0 (
	REM removing quotes if present
	SET options=%nextparam%
	CALL :dequote: options
	IF DEFINED debug ECHO options identified as "!options!"
)
REM p to set port
CALL :param_check p %*
IF ERRORLEVEL 1 SET el=151&&GOTO :endd
IF DEFINED debug ECHO here8
IF %param% GTR 0 (
	SET port=%nextparam%
	IF DEFINED debug ECHO port identified as "!port!"
)

REM b to set basearchive
CALL :param_check b %*
IF ERRORLEVEL 1 SET el=152&&GOTO :endd
IF DEFINED debug ECHO here9
IF %param% GTR 0 (
	SET basearchive=%nextparam%
	FOR %%I IN (!basearchive!) DO SET basearchive=%%~I
	IF DEFINED debug ECHO basearchive identified as "!basearchive!"
)

REM l to set html log file
CALL :param_check l %*
IF ERRORLEVEL 1 SET el=157&&GOTO :endd
IF DEFINED debug ECHO here10
IF %param% GTR 0 (
	REM remove quotes if present
	SET log=%nextparam%
	CALL :dequote log
	IF DEFINED debug ECHO log identified as "!log!"
)

REM e to set excludelist
CALL :param_check e %*
IF ERRORLEVEL 1 SET el=158&&GOTO :endd
IF DEFINED debug ECHO here11
IF %param% GTR 0 (
	REM remove quotes if present
	SET excludelist=%nextparam%
	CALL :dequote excludelist
	IF DEFINED debug ECHO excludelist identified as "!excludelist!"
)

REM i aka lockfree to set VSS
CALL :param_check i %*
IF ERRORLEVEL 1 SET el=153&&GOTO :endd
IF DEFINED debug ECHO here12
IF %param% GTR 0 SET vss=%nextparam%&IF DEFINED debug ECHO vss identified as "!vss!"

REM a to set skip_on_no_change
CALL :param_check a %*
IF ERRORLEVEL 1 SET el=154&&GOTO :endd
IF DEFINED debug ECHO here13
IF %param% GTR 0 SET skip_on_no_change=%nextparam%&IF DEFINED debug ECHO skip_on_no_change identified as "!skip_on_no_change!"

REM j to set skip-rdiff-backup-test
CALL :param_check d %*
IF ERRORLEVEL 1 SET el=156&&GOTO :endd
IF DEFINED debug ECHO here14
IF %param% GTR 0 SET skip-rdiff-backup-test=y&IF DEFINED debug ECHO skip-rdiff-backup-test identified as "!skip-rdiff-backup-test!"

REM t for testing
CALL :param_check t %*
IF ERRORLEVEL 1 SET el=129&&GOTO :endd
IF DEFINED debug ECHO here15
IF %param% GTR 0 (
	SET test=y
	SET action=Test backing up from
	SET testwarn=. Note: no files are backed up, this is JUST A TEST
	IF %quiet% == n ECHO Test run, actual backup will not be performed...
)

REM c for comparing
CALL :param_check c %*
IF ERRORLEVEL 1 SET el=128&&GOTO :endd
IF DEFINED debug ECHO here16
IF %param% GTR 0 (
	SET compare=--compare
	SET action=Comparing
	IF %quiet% == n ECHO Comparing files...
	REM GOTO :done_runtype
)

REM n for command-line specified source,destination pairs - can be specified multiple times
CALL :param_check n %*
IF ERRORLEVEL 1 SET el=159&&GOTO :endd
IF DEFINED debug ECHO here18
IF %param% GTR 0 (
	REM any remaining parameters should be "SourceDirectory,DestinationArchive|..." 
	REM all in quotes, multiple entries separated by vertical stroke
	IF DEFINED debug ECHO Source,dest pairs: %nextparam%
	FOR /F %%A IN (%nextparam%) DO ECHO %%~A >"%TEMP%\%thislc%-%myid%-dataset2.txt"
	"%runfrom%\ssed.exe" -e "s/|/\n/" "%TEMP%\%thislc%-%myid%-dataset2.txt" >"%TEMP%\%thislc%-%myid%-dataset1.txt"
)
IF DEFINED debug ECHO here19
REM completed parsing command line

REM Determine Windows version WINVER 50=2000, 51=XP, 52=2003, 60=Vista, 61=7/2008, 100=10
FOR /F "tokens=2* delims=[]" %%A IN ('VER') DO FOR /F "tokens=2,3 delims=. " %%B IN ("%%A") DO SET WINVER=%%B%%C
REM Set WINBIT as 32-bit (x86) or 64-bit (x64)
SET WINBIT=x86&&IF "%PROCESSOR_ARCHITECTURE%" == "AMD64" (SET WINBIT=x64) ELSE IF "%PROCESSOR_ARCHITEW6432%" == "AMD64" SET WINBIT=x64
IF DEFINED debug ECHO running Windows %WINVER% ^(%WINBIT%^)
IF %WINVER% LSS 51 (
	IF %quiet% == n ECHO Sorry, %this% cannot run under this version of Windows %WINVER%-%WINBIT%.
	SET el=12
	GOTO :endd
)

REM Set VSHADOWVER appropriately for the vshadow-n-[bit].exe programs
SET VSHADOWVER=2008-r2
REM Correct VSHADOWVER for earlier versions of Windows
IF %WINVER%==51 SET VSHADOWVER=xp&&SET WINBIT=x86
IF %WINVER%==52 SET VSHADOWVER=2003&&SET WINBIT=x86
IF %WINVER%==60 SET VSHADOWVER=2008

IF %quiet% == n (
	ECHO Session %datetime% 
	ECHO  performing: %~nx0 %*
)

VERIFY>NUL

REM expand environment variables in dataset1.txt into datasets.txt
CALL :expand "%TEMP%\%thislc%-%myid%-dataset1.txt" "%TEMP%\%thislc%-%myid%-datasets.txt"
IF ERRORLEVEL 1 SET el=140&&GOTO :endd

REM set some defaults before we pick up values from the configuration file
IF DEFINED debug ECHO about to call set_mapdrive, mapdrive is "%mapdrive%"
CALL :set_mapdrive
IF ERRORLEVEL 1 SET el=133&&GOTO :endd
IF DEFINED debug ECHO called set_mapdrive, mapdrive is "%mapdrive%"
SET mainsrcdr=C:
IF ERRORLEVEL 1 SET el=134&&GOTO :endd
IF /I "%forcedeleteshadows%"=="y" SET deleteshadows=y
REM test if we have a server defined
IF NOT DEFINED server ECHO TimeDicer Server not defined&&SET el=15&&GOTO :endd
REM set lastchangesfl and thischangesfl - long filenames without quotes
FOR /F "usebackq delims=" %%I IN (`ECHO "%configfl%"`) DO (
	SET lastchangesfl=%%~dpIlastchanges-%user%-%%~nI.txt
)
SET thischangesfl=%TEMP%\%thislc%-%myid%-thischanges.txt
IF ERRORLEVEL 1 SET el=135&&GOTO :endd

IF DEFINED debug (
	ECHO lastchangesfl set to "%lastchangesfl%"
	ECHO thischangesfl set to "%thischangesfl%"
)
REM remove quotes from 'log'
CALL :dequote log
IF ERRORLEVEL 1 SET el=136&&GOTO :endd
IF DEFINED debug ECHO log is "%log%"
REM remove quotes from 'excludelist', and use shortform named - needed for XCOPY
IF DEFINED excludelist (
	CALL :shortformdequote excludelist
	IF ERRORLEVEL 1 SET el=137&&GOTO :endd
	IF DEFINED debug ECHO excludelist is "%excludelist%"
)
REM remove quotes from 'key' variable
CALL :dequote key
IF ERRORLEVEL 1 SET el=138&&GOTO :endd
IF DEFINED debug ECHO key is "%key%"
REM set default basearchive if not already set
IF NOT DEFINED basearchive SET basearchive=/home/%user%/
REM add trailing slash to basearchive if required
IF "%basearchive:~-1%" NEQ "/" SET basearchive=%basearchive%/
IF ERRORLEVEL 1 SET el=139&&GOTO :endd

REM ---- Test whether the private key file can be found ----
REM The private key in plink.exe is used for authentication - it is not sent to the server, instead
REM a message signed with this private key is sent, and the server uses the public key - already stored in
REM authorized_keys - to verify that the message has been generated using the private key. This proves that the
REM client computer indeed holds this private key and so a connection is granted. The host info is substituted
REM for %s - or %%s in a batch file - i.e. user@server
IF NOT EXIST "%key%" SET el=8&&ECHO Cannot find local key file "%key%"&&GOTO :endd
IF DEFINED debug ECHO Found local Private Key "%key%"
IF "%skip-rdiff-backup-test%"=="y" (
	IF DEFINED debug ECHO Skipped testing for connection to remote server and for /home/tmp
) ELSE (
	REM ---- Test for ssh connection to TimeDicer Server, for write access to our home, and report if /home/tmp exists ----
	SET hometmpexists=x
	REM echoing y ensures that if asked about the public key of the server we accept it
	ECHO y|"%execpath%\plink.exe" -P %port% -ssh -i "%key%" %user%@%server% "echo `[ -d /home/tmp ]&&echo @SET hometmpexists=y>%basearchive%/test.txt||echo @SET hometmpexists=n>%basearchive%/test.txt; cat %basearchive%/test.txt`; rm %basearchive%/test.txt" >%TEMP%\%thislc%-%myid%-testlink.bat
	CALL %TEMP%\%thislc%-%myid%-testlink.bat
	IF !hometmpexists! EQU x (
		SET el=9
		ECHO Cannot make connection to %user%@%server% or no write access to %basearchive%
		GOTO :endd
	)
)
REM if possible use /home/tmp as remote temp, ensures lots of space available for regressions/restores
IF "%hometmpexists%" EQU "y" (SET remotetemp= --remote-temp=/home/tmp) ELSE SET remotetemp= --remote-temp=/tmp
IF DEFINED debug ECHO Successfully connected to Server "%server%" as "%user%", hometmpexists is "%hometmpexists%", remotetemp is "%remotetemp%"
IF ERRORLEVEL 1 SET el=106&&GOTO :endd

REM get version of rdiff-backup.exe
FOR /F "tokens=2* delims= " %%A IN ('%rdiffbackup_on_client% --version') DO FOR /F "tokens=1-3 delims=." %%B IN ("%%A") DO (
	SET RDV1=%%B
	SET RDV2=%%C
	SET RDV3=%%D
)
IF DEFINED debug ECHO We will use %rdiffbackup_on_client% version %RDV1%.%RDV2%.%RDV3%

REM using --test-server does not seem to work for rdiff-backup.exe versions 2.0.0-2.0.3 (at least)
IF %RDV1% GEQ 2 SET skip-rdiff-backup-test=y

REM ---- Test whether we can connect to a compatible rdiff-backup version ----
IF "%skip-rdiff-backup-test%"=="y" (
	IF DEFINED debug ECHO Skipped testing rdiff-backup on remote server
) ELSE (
	IF DEFINED debug (
		ECHO Using Private Key "%key%" and user %user%
		ECHO %rdiffbackup_on_client% --test-server --remote-schema """%execpath%\plink.exe"" -batch -P %port% -ssh -i ""%key%"" %%s %rdiffbackup_on_server% --server" C: %user%@%server%::"%basearchive%"
	)
	%rdiffbackup_on_client% --test-server --remote-schema """%execpath%\plink.exe"" -batch -P %port% -ssh -i ""%key%"" %%s %rdiffbackup_on_server% --server" C: %user%@%server%::"%basearchive%">nul
	IF ERRORLEVEL 1 (
		SET el=14
		ECHO Connected to %user%@%server% but could not make valid connection to rdiff-backup there&&GOTO :endd
	IF DEFINED debug ECHO Found compatible rdiff-backup version at Server %server%
	)
)

REM obtain today's date as variable mdy in mm-dd-yyyy format - locale independent
CALL :today_mdy
IF DEFINED debug ECHO mdy is %mdy%
IF NOT DEFINED mdy ECHO Error obtaining date in mm-dd-yyyy format&&SET el=16&&GOTO :endd
REM check for existence of volume shadowing services, if we might need them
IF ERRORLEVEL 1 SET el=105&&GOTO :endd
IF /I "%vss%" == "y" (
	REM test we have admin permissions
	NET SESSION >NUL 2>&1 || SET el=11&&GOTO :endd
	SET VSHADOW=vshadow-%VSHADOWVER%-%WINBIT%.exe
	IF DEFINED debug ECHO About to check for !VSHADOW!
	IF NOT EXIST !VSHADOW! (
	REM	REM this is untested 2 May 2018 - in particular vscsc.exe may not support -script option
	REM	SET VSHADOW=vscsc.exe
	REM	IF NOT EXIST !VSHADOW! (
			SET el=5&&GOTO :endd
	REM	)
	)
	IF DEFINED debug ECHO Found !VSHADOW! as vshadow program
	IF DEFINED debug ECHO About to check for dosdev.exe
	CALL :file_check dosdev.exe http://www.ltr-data.se/files/dosdev.zip %el% || SET el=5&&GOTO :endd
)
REM Determine mainsrcdr by reference to backup sources, sticking with preset value if it is valid for any
IF DEFINED debug SET /P dummy=Found source drives: <NUL
FOR /F "usebackq tokens=1,2,3* delims=," %%I IN (%TEMP%\%thislc%-%myid%-datasets.txt) DO (
	FOR /F "usebackq tokens=1* delims=:" %%M IN (`ECHO %%I`) DO (
		IF DEFINED debug SET /P dummy= %%M:<NUL
		IF /I x%%M:x==x%mainsrcdr%x (SET srcdone=y) ELSE SET altsource=%%M:
	)
)
IF DEFINED debug ECHO.
IF x%srcdone%x NEQ xyx (
	SET mainsrcdr=%altsource%
	IF DEFINED debug ECHO Altered mainsrcdr to %altsource%
)
REM make sure that mainsrcdr is a single letter and colon
SET mainsrcdr=%mainsrcdr:~0,1%:
IF DEFINED debug ECHO Using VSS mainsrcdr %mainsrcdr%

REM ---- Change current drive ----
REM For reasons that are undocumented - but probably related to the location of
REM snapshot data - vshadow must be run with a local, or the snapshot source,
REM drive as the current drive on the command line. So we must switch to source
REM drive and ensure that all calls to external programs are mapped back to the
REM original location  - which may for instance be on a network share
REM switch to %TEMP% if on mainsrcdr, otherwise to %mainsrcdr%:\
CD %mainsrcdr%\
CD %TEMP%
%mainsrcdr%

REM ---- Test state of Volume Shadow Service Writers ----
REM for info about vshadow see http://msdn.microsoft.com/en-us/library/bb530725%28VS.85%29.aspx
IF DEFINED debug ECHO execpath: "%execpath%"
IF DEFINED debug ECHO startpath: "%startpath%"
IF DEFINED debug ECHO runfrom: "%runfrom%"
IF DEFINED debug ECHO CD: "%CD%"
IF DEFINED debug ECHO Testing for vss=%vss%
IF /I "%vss%" == "y" (
	IF DEFINED debug ECHO "%runfrom%\%VSHADOW%" -ws: starting
	REM allowed status for shadow writers is 1 (stable) or 5 (waiting for completion) - see http://msdn.microsoft.com/en-us/library/aa384979%28VS.85%29.aspx
	REM so count the number of status lines that do not show status 1 or 5...
	"%runfrom%\%VSHADOW%" -ws|"%runfrom%\ssed.exe" -n -e "/Status: [^15]/p"|"%runfrom%\ssed.exe" -n "$=">%TEMP%\%thislc%-%myid%-vsswriters_status.txt
	IF DEFINED debug ECHO "%runfrom%\%VSHADOW%" -ws: completed
	SET VSSNOTREADY=0
	FOR /F "usebackq" %%A IN ("%TEMP%\%thislc%-%myid%-vsswriters_status.txt") DO SET VSSNOTREADY=%%~zA
	IF "!VSSNOTREADY!" NEQ "0" (
		ECHO Volume Shadow Writer[s] not ready, aborting...
		ECHO Check "%TEMP%\%thislc%-%myid%-vsswriters_status.txt"
		SET el=3
		GOTO :endd
	)
	IF ERRORLEVEL 1 SET el=107&&GOTO :endd
	IF DEFINED debug ECHO Volume Shadow Service is available and will be used
) ELSE (
	REM prevent any mapping if vss is off
	SET mapdrive=%mainsrcdr%
	ECHO Volume Shadow Service will not be used
)
IF DEFINED debug ECHO locating an available drive letter DR
CALL :finddrive
IF NOT DEFINED DR SET DR=C:\
IF DEFINED debug ECHO DR set to !DR!
REM ---- Process excludelist to adapt from Windows to Linux format ----
IF DEFINED debug ECHO About to process excludelist to adapt to Windows format
IF "x%excludelist%" NEQ "x" (
	IF NOT EXIST "%excludelist%" SET el=4&&ECHO Cannot find excludelist file "%excludelist%"&&GOTO :endd
	REM expand env variables in excludelist
	CALL :expand "%excludelist%" "%TEMP%\%thislc%-%myid%-excludelist1.txt"
	REM set exclude filename/path in linuxed form, note the additional space at the end
	SET exclude=--exclude-globbing-filelist %TEMP:\=/%/%thislc%-%myid%-excludelist.txt 
	REM create exclude file so that it holds data in suitable linuxed form
	REM /^^#/d - delete: exclude comment lines
	REM /^^$/d - delete: exclude blank lines
	REM s/^%mainsrcdr%/%mapdrive%/i - change refs from C: to mapped drive
	REM /^[A-Za-z]:/!s/^/**/; - prefix ** but only if it doesn't start with drive letter
	REM s/^/ignorecase:/ - prefix ignorecase:
	REM s/\\/\//g - substitute: switch any \ to / - and g means repeat action
	REM s/$/**/ - substitute: suffix **
	SETLOCAL ENABLEEXTENSIONS DISABLEDELAYEDEXPANSION
	REM IF %RDV1% GEQ 2 (
	REM 	REM version for rdiff-backup v2+:
	REM 	REM this does not work: ignorecase:**\\/Rents-0.vmdk
	REM 	IF DEFINED debug ECHO Doing ssed.exe "/^#/d;/^$/d;s/^%mainsrcdr%/%mapdrive%/i;/^[A-Za-z]:/!s/^/**/;s/^/ignorecase:/;s/\\/\\\\\//g;s/$/**/" "%TEMP%\%thislc%-%myid%-excludelist1.txt" ^>"%TEMP%\%thislc%-%myid%-excludelist.txt"
	REM 	"%runfrom%\ssed.exe" "/^#/d;/^$/d;s/^%mainsrcdr%/%mapdrive%/i;s/^/ignorecase:**/;s/\\/\\\\\//g;s/$/**/" "%TEMP%\%thislc%-%myid%-excludelist1.txt" >"%TEMP%\%thislc%-%myid%-excludelist.txt"
	REM ) ELSE (
		REM version for rdiff-backup v1.2.8 or v1.3.3:
		IF DEFINED debug ECHO Doing ssed.exe "/^#/d;/^$/d;s/^%mainsrcdr%/%mapdrive%/i;/^[A-Za-z]:/!s/^/**/;s/^/ignorecase:/;s/\\/\//g;s/$/**/" "%TEMP%\%thislc%-%myid%-excludelist1.txt" ^>"%TEMP%\%thislc%-%myid%-excludelist.txt"
		"%runfrom%\ssed.exe" "/^#/d;/^$/d;s/^%mainsrcdr%/%mapdrive%/i;s/^/ignorecase:**/;s/\\/\//g;s/$/**/" "%TEMP%\%thislc%-%myid%-excludelist1.txt" >"%TEMP%\%thislc%-%myid%-excludelist.txt"
	REM )
	SETLOCAL ENABLEEXTENSIONS ENABLEDELAYEDEXPANSION
)
IF ERRORLEVEL 1 SET el=114&&ECHO Error using ssed to create %TEMP%\%thislc%-%myid%-excludelist.txt&&GOTO :endd

IF NOT EXIST "%MYAPPDATA%" MKDIR "%MYAPPDATA%" 2>nul
IF ERRORLEVEL 1 SET el=108&&ECHO Error creating %MYAPPDATA%&&GOTO :endd
REM ---- %thislc%-%myid%-action.bat script creation ----
REM This script will shortly be run by vshadow
IF DEFINED debug ECHO Start creating script %TEMP%\%thislc%-%myid%-action.bat
SETLOCAL ENABLEEXTENSIONS DISABLEDELAYEDEXPANSION
ECHO @ECHO OFF>%TEMP%\%thislc%-%myid%-action.bat
ECHO SETLOCAL ENABLEEXTENSIONS ENABLEDELAYEDEXPANSION>>%TEMP%\%thislc%-%myid%-action.bat
IF /I x%vss% == xy (
	ECHO REM map the shadow device to drive %mapdrive%>>%TEMP%\%thislc%-%myid%-action.bat
	ECHO CALL %TEMP%\%thislc%-%myid%-vss-setvar.cmd>>%TEMP%\%thislc%-%myid%-action.bat
	ECHO "%runfrom%\dosdev" %mapdrive% %%SHADOW_DEVICE_1%% 2^>nul>>%TEMP%\%thislc%-%myid%-action.bat
)
ECHO FOR /F "usebackq tokens=1,2,3* delims=," %%%%I IN ^(%TEMP%\%thislc%-%myid%-datasets.txt^) DO ^(>>%TEMP%\%thislc%-%myid%-action.bat
ECHO 	REM must set non-blank addparam here to prevent ERRORLEVEL 1 when it is set to blank>>%TEMP%\%thislc%-%myid%-action.bat
ECHO 	SET addparam=x>>%TEMP%\%thislc%-%myid%-action.bat
REM 	unixify source [first field:%%I] on current line in list file, fix any drive ref and set as !source!, also unixify parameters [%%K]
ECHO 	FOR /F "usebackq tokens=1,2* delims=�" %%%%M IN ^(`ECHO %%%%I�%%%%K^^^|"%runfrom%\ssed.exe" -e s/%mainsrcdr%/%mapdrive%/gi -e s@\\@/@g`^) DO SET source=%%%%M^&IF x%%%%Kx NEQ xx ^(SET addparam=%%%%N^) ELSE SET addparam=>>%TEMP%\%thislc%-%myid%-action.bat
IF DEFINED debug ECHO 	ECHO vars I,J,K,M,N are: %%%%I,%%%%J,%%%%K,%%%%M,%%%%N^>CON>>%TEMP%\%thislc%-%myid%-action.bat
REM IF /I "%vss%" == "y" (
REM 	ECHO 	REM For xcopy, set xcopyto to expanded var I mapped to VSS drive>>%TEMP%\%thislc%-%myid%-action.bat
REM 	ECHO 	FOR /F "usebackq tokens=1* delims=?" %%%%W IN ^(`ECHO %%%%I^^^|"%runfrom%\ssed.exe" -e s/%mainsrcdr%/%mapdrive%/gi`^) DO IF /I %%%%W==%%%%I ^(SET xcopyto=%%%%~dW^) ELSE SET xcopyto=%%%%W\>>%TEMP%\%thislc%-%myid%-action.bat
REM ) ELSE (
ECHO 	REM xcopy will fail if the only available drive is C:\>>%TEMP%\%thislc%-%myid%-action.bat
ECHO 	REM and C:\ is itself a backup source>>%TEMP%\%thislc%-%myid%-action.bat
ECHO  	IF /I "%%%%I"=="C:\" ^(SET xcopyto=%DR%^) ELSE SET xcopyto=C:\>>%TEMP%\%thislc%-%myid%-action.bat
REM )
IF DEFINED debug ECHO 	ECHO xcopyto is !xcopyto!^>CON>>%TEMP%\%thislc%-%myid%-action.bat
ECHO 	IF NOT EXIST "!source!" ECHO %%%%I - skipped as non-existent>>%TEMP%\%thislc%-%myid%-action.bat
ECHO 	IF EXIST "!source!" ^(>>%TEMP%\%thislc%-%myid%-action.bat
IF %quiet% == n ECHO 		ECHO %action% %%%%I to %%%%J...^>CON>>%TEMP%\%thislc%-%myid%-action.bat
ECHO 		SET fcount=999>>%TEMP%\%thislc%-%myid%-action.bat
ECHO 		REM decide whether to skip the checking of whether files have changed in this archive since last run>>%TEMP%\%thislc%-%myid%-action.bat
ECHO 		SET skipchk=N>>%TEMP%\%thislc%-%myid%-action.bat
ECHO 		REM skip checking if doing a compare run>>%TEMP%\%thislc%-%myid%-action.bat
ECHO 		IF DEFINED compare SET skipchk=Y>>%TEMP%\%thislc%-%myid%-action.bat
ECHO 		REM skip checking if skip_on_no_change is not y>>%TEMP%\%thislc%-%myid%-action.bat
ECHO 		IF /I "%skip_on_no_change%" NEQ "y" SET skipchk=Y>>%TEMP%\%thislc%-%myid%-action.bat
ECHO 		REM skip checking if there are additional parameters cos xcopy cant reliably identify whether files have changed>>%TEMP%\%thislc%-%myid%-action.bat
ECHO 		IF DEFINED K SET skipchk=Y>>%TEMP%\%thislc%-%myid%-action.bat
REM ECHO 		ECHO skipchk is ^!skipchk^!, skip_on_no_change is %skip_on_no_change%^>CON>>%TEMP%\%thislc%-%myid%-action.bat
ECHO 		REM skip checking if no file recording last successful update>>%TEMP%\%thislc%-%myid%-action.bat
ECHO 		IF NOT EXIST "%lastchangesfl%" SET skipchk=Y>>%TEMP%\%thislc%-%myid%-action.bat
ECHO 		IF "!skipchk!"=="N" (>>%TEMP%\%thislc%-%myid%-action.bat
REM IF %quiet% == n ECHO 			ECHO Checking to see if any files have changed since last backup...^>CON>>%TEMP%\%thislc%-%myid%-action.bat
ECHO 			REM now examine %lastchangesfl%>>%TEMP%\%thislc%-%myid%-action.bat
ECHO 			SET datasetlastchanged=[never]>>%TEMP%\%thislc%-%myid%-action.bat
ECHO 			FOR /F "usebackq tokens=1,2* delims=," %%%%R IN ^("%lastchangesfl%"^) DO ^(>>%TEMP%\%thislc%-%myid%-action.bat
ECHO 				IF /I "%%%%R,%%%%S"=="%%%%I,%%%%J" ^(>>%TEMP%\%thislc%-%myid%-action.bat
ECHO 					REM if found matching entry in %lastchangesfl% use xcopy to check for changes to it on or after the last backup date>>%TEMP%\%thislc%-%myid%-action.bat
IF DEFINED debug 					ECHO Doing: xcopy ^^^"%%%%I\*^^^" ^^^"!xcopyto!^^^" /d:%%%%T /leqry /exclude:%excludelist%
ECHO 					FOR /F "tokens=1" %%%%U IN ^('xcopy ^^^"%%%%I\*^^^" ^^^"!xcopyto!^^^" /d:%%%%T /leqry /exclude:%excludelist%'^) DO SET fcount=%%%%U>>%TEMP%\%thislc%-%myid%-action.bat
ECHO 					IF ERRORLEVEL 1 SET fcount=999>>%TEMP%\%thislc%-%myid%-action.bat
ECHO 					REM this records the previous copy date in the current this_success_date file because there will be no update>>%TEMP%\%thislc%-%myid%-action.bat
ECHO 					IF !fcount!==0 ECHO %%%%I,%%%%J,%%%%T^>^>"%thischangesfl%">>%TEMP%\%thislc%-%myid%-action.bat
ECHO 					SET datasetlastchanged=%%%%T>>%TEMP%\%thislc%-%myid%-action.bat
ECHO 				^)>>%TEMP%\%thislc%-%myid%-action.bat
ECHO 			^)>>%TEMP%\%thislc%-%myid%-action.bat
ECHO 		^)>>%TEMP%\%thislc%-%myid%-action.bat
ECHO 		IF !fcount! == 0 ^(>>%TEMP%\%thislc%-%myid%-action.bat
ECHO 			ECHO %%%%I - skipped as unchanged since m-d-y !datasetlastchanged! 00:00>>%TEMP%\%thislc%-%myid%-action.bat
IF %quiet% == n ECHO 			ECHO   skipped as unchanged since m-d-y !datasetlastchanged! 00:00^>CON>>%TEMP%\%thislc%-%myid%-action.bat
ECHO 		^) ELSE ^(>>%TEMP%\%thislc%-%myid%-action.bat
IF %quiet% == n ECHO 			IF !fcount! NEQ 999 ECHO   !fcount! new/changed files since m-d-y !datasetlastchanged! 00:00, backup is proceeding...^>CON>>%TEMP%\%thislc%-%myid%-action.bat
IF %quiet% == n ECHO 			IF !fcount! EQU 999 IF /I "%skip_on_no_change%" == "y" ECHO   unable to confirm that no files have changed, backup is proceeding...^>CON>>%TEMP%\%thislc%-%myid%-action.bat
ECHO 			ECHO ^^^<a href="#%datetime%-%%%%J"^^^>%%%%I to %%%%J^^^</a^^^> >>%TEMP%\%thislc%-%myid%-action.bat
ECHO 			IF x!rdiffhasrun!==x ECHO ^^^<h3^^^>Details^^^</h3^^^>^>^>%log2%>>%TEMP%\%thislc%-%myid%-action.bat
ECHO 			SET rdiffhasrun=y>>%TEMP%\%thislc%-%myid%-action.bat
ECHO 			ECHO ^^^<h4^^^>^^^<a name="%datetime%-%%%%J"^^^>^^^</a^^^>%action% %%%%I to %%%%J^^^</h4^^^>^>^>%log2%>>%TEMP%\%thislc%-%myid%-action.bat
ECHO 			IF !fcount! EQU 999 IF /I "%skip_on_no_change%" == "y" ECHO Unable to confirm that no files have changed, so backup is proceeding...^>^>%log2%>>%TEMP%\%thislc%-%myid%-action.bat
ECHO 			ECHO %action% "!source!" with command: "%runfrom%\%rdiffbackup_on_client%" %compare% --create-full-path --no-hard-links --print-statistics --remote-schema """%execpath%\plink.exe"" -batch -P %port% -ssh -i ""%key%"" %%%%s %rdiffbackup_on_server% --server" !addparam!%remotetemp% %options% %exclude%"!source!" %user%@%server%::"%basearchive%%%%%J"^>^>%log2% 2^>^>^&^1>>%TEMP%\%thislc%-%myid%-action.bat
IF NOT DEFINED test ECHO 			"%runfrom%\%rdiffbackup_on_client%" %compare% --create-full-path --no-hard-links --print-statistics --remote-schema """%execpath%\plink.exe"" -batch -P %port% -ssh -i ""%key%"" %%%%s %rdiffbackup_on_server% --server" !addparam!%remotetemp% %options% %exclude%"!source!" %user%@%server%::"%basearchive%%%%%J"^>^>%log2% 2^>^>^&^1>>%TEMP%\%thislc%-%myid%-action.bat
ECHO 			SET ACTERR=!ERRORLEVEL!>>%TEMP%\%thislc%-%myid%-action.bat
ECHO 			IF !ACTERR!==0 ECHO %%%%I,%%%%J,%mdy%^>^>"%thischangesfl%" ^2^>^>^&^1 >>%TEMP%\%thislc%-%myid%-action.bat
IF DEFINED debug ECHO 			ECHO ACTERR is !ACTERR!^>CON>>%TEMP%\%thislc%-%myid%-action.bat

ECHO 			REM Pause until we can write to %log2% without error>>%TEMP%\%thislc%-%myid%-action.bat
IF DEFINED debug ECHO 			ECHO About to enter NETSH loop^>CON>>%TEMP%\%thislc%-%myid%-action.bat
ECHO 			CALL :echotest>>%TEMP%\%thislc%-%myid%-action.bat
ECHO 			REM Clear ERRORLEVEL that may be set by the NETSH^>NUL>>%TEMP%\%thislc%-%myid%-action.bat
ECHO 			VERIFY^>NUL>>%TEMP%\%thislc%-%myid%-action.bat
ECHO 			IF !ACTERR! GTR 0 ^(>>%TEMP%\%thislc%-%myid%-action.bat
ECHO 				ECHO Error 30 - internal rdiff-backup error !ACTERR! - occurred with %action% %%%%I to %%%%J^>^>%log2%>>%TEMP%\%thislc%-%myid%-action.bat
IF /I "%vss%" == "y" (
	ECHO 				ECHO   Error 30 - internal rdiff-backup error !ACTERR! - occurred with %action% %%%%I to %%%%J^>CON>>%TEMP%\%thislc%-%myid%-action.bat
)
ECHO 				ECHO SET ACTERR=30^>%TEMP%\%thislc%-%myid%-action-err.bat>>%TEMP%\%thislc%-%myid%-action.bat
ECHO 			^)>>%TEMP%\%thislc%-%myid%-action.bat
ECHO 			ECHO %action% %%%%I to %%%%J completed !DATE! !TIME!%testwarn%.^>^>%log2%>>%TEMP%\%thislc%-%myid%-action.bat
ECHO 			ECHO ^^^<a href="#%datetime%"^^^>Return to top of session %datetime%^^^</a^^^> ^>^>%log2%>>%TEMP%\%thislc%-%myid%-action.bat
ECHO 		^)>>%TEMP%\%thislc%-%myid%-action.bat
IF DEFINED debug ECHO 		ECHO About to write to %log2%^>CON>>%TEMP%\%thislc%-%myid%-action.bat
REM IF %quiet% == n ECHO 		ECHO.^>CON>>%TEMP%\%thislc%-%myid%-action.bat
ECHO 	^)>>%TEMP%\%thislc%-%myid%-action.bat
ECHO ^)>>%TEMP%\%thislc%-%myid%-action.bat
IF /I "%vss%" == "y" (
	ECHO REM delete shadow device drive mapping>>%TEMP%\%thislc%-%myid%-action.bat
	ECHO "%runfrom%\dosdev" -r -d %mapdrive% 2^>NUL>>%TEMP%\%thislc%-%myid%-action.bat
)
ECHO IF NOT DEFINED fcount ECHO SET ACTERR=19^>%TEMP%\%thislc%-%myid%-action-err.bat>>%TEMP%\%thislc%-%myid%-action.bat
ECHO GOTO :EOF>>%TEMP%\%thislc%-%myid%-action.bat
ECHO :echotest>>%TEMP%\%thislc%-%myid%-action.bat
ECHO DEL %TEMP%\%thislc%-%myid%-echoerr.txt ^>NUL ^2^>^&^1 >>%TEMP%\%thislc%-%myid%-action.bat
ECHO SET /P dummy=.^<NUL^>^>%log2% ^2^>%TEMP%\%thislc%-%myid%-echoerr.txt>>%TEMP%\%thislc%-%myid%-action.bat
ECHO fc %TEMP%\%thislc%-%myid%-echoerr.txt %TEMP%\%thislc%-%myid%-nul.txt^>NUL ^2^>^&^1 >>%TEMP%\%thislc%-%myid%-action.bat
ECHO IF !ERRORLEVEL!==0 GOTO :EOF>>%TEMP%\%thislc%-%myid%-action.bat
IF DEFINED debug ECHO SET /P dummy=.^<NUL^>CON>>%TEMP%\%thislc%-%myid%-action.bat
ECHO NETSH Diag Ping Loopback^>NUL ^2^>^&^1 >>%TEMP%\%thislc%-%myid%-action.bat
ECHO GOTO :echotest>>%TEMP%\%thislc%-%myid%-action.bat
ENDLOCAL
IF DEFINED debug ECHO End creating script %TEMP%\%thislc%-%myid%-action.bat
REM ---- Tidy up before starting the volume shadowing and backup ----
REM delete any existing shadow copies  - there should not normally be any, but can be if a previous backup failed
IF /I x%vss% == xy (
	IF ERRORLEVEL 1 SET el=109&&GOTO :endd
	IF NOT "%deleteshadows%"=="y" (
		IF DEFINED debug ECHO Skipping deletion of any existing shadow copies
	) ELSE (
		IF DEFINED debug ECHO About to delete any existing shadow copies
		ECHO y|"%runfrom%\%VSHADOW%" -da>nul || SET el=7&&GOTO :endd
		IF DEFINED debug ECHO Deleted any existing shadow copies
	)
)

REM delete any this_success_date file - should not be one unless a previous run ended prematurely
IF EXIST "%thischangesfl%" (
	DEL "%thischangesfl%">nul 2>nul
	IF ERRORLEVEL 1 SET el=101&&ECHO Error deleting old success date file&&GOTO :endd
)

IF DEFINED debug (
	ECHO About to do backup run
	IF "%exitnow%" NEQ "y" PAUSE
)
REM delete any existing log files
IF EXIST "%log2%" DEL "%log2%" 2>nul
IF ERRORLEVEL 1 SET el=110&&ECHO Error deleting old log file %log2%&&GOTO :endd
REM create a null file for comparison purposes
TYPE NUL>%TEMP%\%thislc%-%myid%-nul.txt
REM ---- Do the backup ----
IF EXIST %TEMP%\%thislc%-%myid%-action-err.bat DEL %TEMP%\%thislc%-%myid%-action-err.bat>NUL
SET ACTERR=0
IF %quiet% == n ECHO %action% selected locations is now commencing
REM ECHO ^<p^>%action% %mainsrcdr% ^(as %mapdrive%^) using %~nx0 v%ver% with configuration "%configfl%".^</p^>>>%log1%
ECHO ^<p^>%action% %mainsrcdr% ^(as %mapdrive%^) using %this% v%ver%: %~nx0 %*%testwarn%.^</p^>>>%log1%
ECHO ^<h3^>Summary^</h3^>>>%log1%
IF /I x%vss% == xy (
	IF ERRORLEVEL 1 SET el=111&&GOTO :endd
	REM ---- Run vshadow, which will create shadow copy, run %thislc%-%myid%-action.bat, then delete shadow copy ----
	IF DEFINED debug ECHO About to run "%thislc%-%myid%-action.bat" in VSS mode
	"%runfrom%\%VSHADOW%" -script=%TEMP%\%thislc%-%myid%-vss-setvar.cmd -exec=%TEMP%\%thislc%-%myid%-action.bat %mainsrcdr%>>"%log1%" 2>>&1
	IF ERRORLEVEL 1 SET /A el=%ERRORLEVEL%+20
	IF DEFINED debug ECHO Returned from running "%thislc%-%myid%-action.bat" in VSS mode
) ELSE (
	IF ERRORLEVEL 1 SET el=112&&GOTO :endd
	IF DEFINED debug ECHO About to run "%thislc%-%myid%-action.bat" in non-VSS mode
	CALL %TEMP%\%thislc%-%myid%-action.bat>>"%log1%" 2>>&1
	IF ERRORLEVEL 1 SET el=%ERRORLEVEL%
	IF DEFINED debug ECHO Returned from running "%thislc%-%myid%-action.bat" in non-VSS mode
)
ECHO Session %datetime% completed %DATE% %TIME%>>%log1%
IF EXIST %TEMP%\%thislc%-%myid%-action-err.bat CALL %TEMP%\%thislc%-%myid%-action-err.bat
IF DEFINED debug ECHO ACTERR is "%ACTERR%"
SET /A el=%el%+%ACTERR%

REM Clear the ERRORLEVEL
VERIFY>NUL
IF %quiet% == n ECHO %action% selected locations completed

REM ---- Tidy up after backup run ----
IF DEFINED debug ECHO el is "%el%"
IF EXIST %log1% (
	IF ERRORLEVEL 1 SET el=113&&GOTO :endd	
	IF DEFINED debug ECHO About to modify "%log1%"
	COPY /Y /A %log1% %log1%x>NUL 2>&1
	REM terminate all lines with '<br />' unless already ending in a '>'
	IF "%el%" == "0" (
		REM if no errors, remove all the output guff produced by VSHADOW.
		"%runfrom%\ssed.exe" "/This script is generated by VSHADOW.EXE/{N;d};/^VSHADOW.EXE/,/Snapshot creation done/d;/The process cannot access the file because it is being used by another process./d;s@\([^>]\)$@\1<br />@" %log1%x>%log1%
	) ELSE (
		"%runfrom%\ssed.exe" "/This script is generated by VSHADOW.EXE/{N;d};/The process cannot access the file because it is being used by another process./d;s@\([^>]\)$@\1<br />@" %log1%x>%log1%
	)
	IF ERRORLEVEL 1 SET el=102&&GOTO :endd
	IF DEFINED debug ECHO Modified "%log1%"
)
IF DEFINED debug ECHO About to combine logs
REM Build log at %log1%y
IF ERRORLEVEL 1 SET el=141&&GOTO :endd
REM This is the top of the log file
REM Echo escaping rules from http://www.robvanderwoude.com/escapechars.php:
REM [&<>|] 	prefix ^	always
REM ^		prefix ^	not always required in doublequoted strings, but won't hurt
REM %		prefix %	not always required in doublequoted strings, just try
REM !		prefix ^!	only when delayed variable expansion is active

ECHO ^<^^!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Transitional//EN" "http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd"^>>%log1%y
ECHO ^<html xmlns="http://www.w3.org/1999/xhtml"^>>>%log1%y
ECHO ^<head^>>>%log1%y
ECHO ^<link rel="stylesheet" href="timedice.css" /^>>>%log1%y
ECHO ^<link rel="icon" href="timedicer64x64.ico" /^>>>%log1%y
ECHO ^<title^>TimeDicer Log for ^&lsquo;ZuserZ^&rsquo;^</title^>>>%log1%y
ECHO ^</head^>>>%log1%y
ECHO ^<body^>>>%log1%y
ECHO ^<div id="leftcontent"^>>>%log1%y
ECHO ^<a title="Top of Page" href=""^>^<img hspace="10" src="timedicer_dice.png" width="95" height="70" align="center" /^>>>%log1%y
ECHO ^<img src="timedicer_word_small.png" alt="TimeDicer" width="110" height="18" align="center" /^>^</a^>>>%log1%y
ECHO ^<h3^>Links^</h3^>>>%log1%y
ECHO ^<p class=closer^>^<a href="https://www.timedicer.co.uk/"^>TimeDicer Website^</a^>^<br /^>>>%log1%y
ECHO ^<a href="https://www.timedicer.co.uk/timedicer-man.html"^>TimeDicer Manual^</a^>^</p^>>>%log1%y
ECHO ^<h3^>Session List^</h3^>>>%log1%y
ECHO ^<p class="closer"^>^<a href="#ZdatetimeZ"^>ZdatetimeZ^</a^>^</p^>>>%log1%y

REM This is the top of the log file
REM GOTO :step2
REM <!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Transitional//EN" "http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd">
REM <html xmlns="http://www.w3.org/1999/xhtml">
REM <head>
REM <link rel="stylesheet" href="timedice.css" />
REM <link rel="icon" href="timedicer64x64.ico" />
REM <title>TimeDicer Log for &lsquo;ZuserZ&rsquo;</title>
REM </head>
REM <body>
REM <div id="leftcontent">
REM <a title="Top of Page" href=""><img hspace="10" src="timedicer_dice.png" width="95" height="70" align="center" />
REM <img src="timedicer_word_small.png" alt="TimeDicer" width="110" height="18" align="center" /></a>
REM <h3>Links</h3>
REM <p class=closer><a href="https://www.timedicer.co.uk/">TimeDicer Website</a><br />
REM <a href="https://www.timedicer.co.uk/timedicer-man.html">TimeDicer Manual</a></p>
REM <h3>Session List</h3>
REM <p class="closer"><a href="#ZdatetimeZ">ZdatetimeZ</a></p>
REM :step2
:endd
REM this line extracts the lines above to the top of the log file
REM "%runfrom%\ssed.exe" -n "/GOTO :step2/,${/GOTO :step2/d;/^:step2/q;p;}" "%~f0">%log1%y
REM add existing menu items in log file
IF EXIST "%log%" "%runfrom%\ssed.exe" -n "/<p class=\"closer\"/p" "%log%">>%log1%y
REM end of the menu
ECHO ^</div^>>>%log1%y
REM start of the main content
ECHO ^<div id="maincontent"^>^<div class="innertube"^>>>%log1%y
ECHO ^<h1^>^<img src="timedicer_full.png" alt="TimeDicer" /^>^<br /^>^<br /^>Log for ^&lsquo;%user%^&rsquo;^</h1^>^<br /^>>>%log1%y
REM add in text from log1
IF EXIST %log1% TYPE %log1%>>%log1%y
IF ERRORLEVEL 1 IF %el% EQU 0 SET el=142&&GOTO :endd
REM add in text from log2, with line breaks
IF EXIST %log2% "%runfrom%\ssed.exe" "s@\([^>]\)$@\1<br />@" %log2%>>%log1%y
IF ERRORLEVEL 1 IF %el% EQU 0 SET el=143&&GOTO :endd
REM add in text from pre-existing log
IF EXIST "%log%" "%runfrom%\ssed.exe" "/DOCTYPE html PUBLIC/,/<\/h1>/d;/<\/div><\/div><\/body>/d" "%log%">>%log1%y
ECHO ^</div^>^</div^>^</body^>>>%log1%y
IF ERRORLEVEL 1 IF %el% EQU 0 SET el=144&&GOTO :endd
REM copy completed log file to permanent location, and fix variables
IF DEFINED log "%runfrom%\ssed.exe" "s@ZuserZ@%user%@g;s@ZdatetimeZ@%datetime%@g" %log1%y>"%log%"
IF ERRORLEVEL 1 IF %el% EQU 0 SET el=145&&GOTO :endd
IF DEFINED log SET logbuilt=y

REM ---- Ending ----
REM if doing a compare run then show the result
IF DEFINED debug ECHO Reached endd stage
IF DEFINED debug ECHO "%%compare%%1%%test%%2" is "%compare%1%test%2"
IF DEFINED debug ECHO el is "%el%"
IF "%compare%1%test%2" == "12" (
	IF EXIST "%thischangesfl%" (
		IF DEFINED debug ECHO Found "%thischangesfl%"
		REM any datasets missing from this e.g. temporarily edited out but found in last are added back to this
		IF EXIST "%lastchangesfl%" (
			IF DEFINED debug ECHO About to add back missing datasets in "%thischangesfl%"
			FOR /F "usebackq tokens=1-3* delims=," %%I IN ("%lastchangesfl%") DO (
				SET FOUND=N
				FOR /F "usebackq tokens=1-3* delims=," %%L IN ("%thischangesfl%") DO (
					IF /I %%I%%J==%%L%%M SET FOUND=Y
					IF DEFINED debug ECHO Compared last %%I%%J with this %%L%%M - match !FOUND!
				)
				IF ERRORLEVEL 1 IF %el% EQU 0 SET el=115&&GOTO :endd
				REM ECHO %%J !FOUND! FOUND
				IF !FOUND!==N ECHO %%I,%%J,%%K>>"%thischangesfl%"
			)
		)
		IF DEFINED debug ECHO About to replace "%lastchangesfl%" with "%thischangesfl%"
		COPY /Y "%thischangesfl%" "%lastchangesfl%">NUL
		IF ERRORLEVEL 1 IF %el% EQU 0 SET el=117&&GOTO :endd
		IF DEFINED debug ECHO Copied "%thischangesfl%" to "%lastchangesfl%"
	)
)

:end1
IF %quiet% == n (
	IF %el% GEQ 1 (
		CALL :errortext %el%
		REM new code here 12/12/11 - largely untested
		SET FULLERRORTEXT=%~n0 failed with error %el% - !ERRORTEXT!
		IF DEFINED debug ECHO At stage 1, FULLERRORTEXT is !FULLERRORTEXT! and el is %el%
		IF "%ACTERR%" EQU "30" (
			SET FULLERRORTEXT=!FULLERRORTEXT! relating to dataset[s]
			IF DEFINED debug ECHO At stage 2, FULLERRORTEXT is !FULLERRORTEXT!
			"%runfrom%\ssed.exe" -n "/Error 30/s/.*to \(.*\)<.*/\1/p;/Return to top of session/q" "%log%">"%log1%z"
			IF EXIST "%log1%z" FOR /F %%i IN (%log1%z) DO SET FULLERRORTEXT=!FULLERRORTEXT! %%i
			IF DEFINED debug ECHO At stage 3, FULLERRORTEXT is !FULLERRORTEXT!
		)
		ECHO !FULLERRORTEXT!
		IF x%logbuilt%x NEQ xx (
			REM update the log with info about the error
			COPY /Y "%log%" %log1%y>NUL
			"%runfrom%\ssed.exe" "s@\(Session %datetime%</h[0-9]>\)@\1\r\n<p><span class=\"highlight\">!FULLERRORTEXT!</span></p>@" %log1%y>"%log%"
		)
	) ELSE (
		IF ERRORLEVEL 1 (
			REM would this code ever run?
			ECHO %~n0 v%ver% failed with Errorlevel %ERRORLEVEL%
			SET el=%ERRORLEVEL%
			IF %el% == 2 ECHO Check if drive %mainsrcdr% is formatted with FAT32. If so, convert it to NTFS or add&&ECHO SET vss=n to your configuration file
			ECHO For more info run %~n0 /h, or visit %homepage%/timedicer-man.html
		) ELSE (
			ECHO %~n0 completed without errors at %DATE% %TIME%
			REM save the current date so it can be picked up next time
			REM ECHO %mdy%>"%thislc%-%myid%-%user%-last_success_date.txt"
		)
	)
)
IF %quiet% == n (IF x%logbuilt%x NEQ xx ECHO Log at %log%)
REM Restore the original command line path
IF DEFINED startpath CD /D %startpath%
REM view the log if there is one & there were errors or it is compare run
IF x%logbuilt%x == xx GOTO :skipshowlog
IF %el%x%compare% == 0x GOTO :skipshowlog
REM start /MIN cmd /c "%log%"
CALL :openhtml "%log%"
IF NOT "%exitnow%" == "y" PAUSE
:skipshowlog
REM undocumented option 'SET retaintempfiles=Y' retains temporary files
IF NOT DEFINED debug IF %el% EQU 0 IF /I NOT "%retaintempfiles%" == "Y" DEL /Q "%TEMP%\%thislc%*.*" 2>NUL
IF DEFINED debug ECHO Exiting program with code %el%
EXIT /B %el%
REM ---- Ended ----

:set_mapdrive
REM A or B can give 'drive not ready' messages, and we ignore C because it is never available, right?
REM on some systems 'drive not ready' messages occur with other drive letters D-G, prob empty DVD drives etc.
REM the strategy here is first to try W->P, then X->Z, then mid-letters O->I. Surely one will be available?
PUSHD & FOR %%i IN (W V U T S R Q P Z Y X O N M L K J I) DO ( %%i 2>nul || ( SET mapdrive=%%i:&POPD&EXIT /B 0 ) )
REM if we got here then X: shouldn't work, but let's try it anyway
SET mapdrive=X:
GOTO :EOF
:escaped
:: Subroutine to double-escape a string so C:\Folder becomes C:\\Folder etc
IF "%1" EQU "" GOTO :EOF
FOR %%i IN ("\=\\") DO CALL SET "%1=%%%1:%%~i%%"
GOTO :EOF
:expand
:: Subroutine to expand environment variables in a file
::  2 params: 1st is source file, 2nd is dest file
REM IF DEFINED debug ECHO expanding "%~1" into "%~2"
IF EXIST "%~2" DEL "%~2"
SET expandto=%~2
FOR /F "delims=" %%i IN (%~1) DO (
	REM IF DEFINED debug ECHO original line data: "%%i"
	CALL :expand1 %%i
)
REM if source file is nul then ensure expanded file is nul, not missing
IF NOT EXIST "%expandto%" TYPE NUL >"%expandto%"
GOTO :EOF
:expand1
REM %* doesnt work here when compiled with Abyssmedia Quick Batch File Compiler 3.6.1.0, try %~1
ECHO %*>>"%expandto%"
REM IF DEFINED debug ECHO expanded line data: "%~1"
GOTO :EOF
:finddrive
REM Set DR to an available probably writeable non-C drive
FOR /f "tokens=*" %%a IN ('fsutil fsinfo drives^|%findprog% /V ""') DO (
	FOR %%b IN (%%a) DO (
		IF NOT "%%b" == "Drives:" (
			FOR /f "tokens=1-3" %%c IN ('fsutil fsinfo drivetype %%b') DO (
				IF NOT "%%e" == "CD-ROM" IF NOT "%%c" == "A:\" IF NOT "%%c" == "B:\" IF NOT "%%c" == "C:\" SET DR=%%c
			)
		)
	)
)
GOTO :eof
:file_check
IF NOT EXIST "%~1" (
	SET /A el=%3+1
	ECHO  	%1 could not be found: get it from %~2
)
GOTO :eof
:dequote
REM remove outer quotes - example (for %myvar%): CALL :dequote myvar
FOR /f "delims=" %%A in ('echo %%%1%%') DO SET %1=%%~A
GOTO :eof
:shortformdequote
REM remove outer quotes and render in short filename form - example (for var %myvar%): CALL :shortformdequote myvar
FOR /F "delims=" %%A in ('ECHO %%%1%%') DO SET %1=%%~sA
GOTO :eof
:param_check
REM call as e.g. CALL :param_check q %* where /q or -q is the param looked for
REM (not case sensitive). If param is found then on return param holds the parameter position, o/wise 0
REM the next parameter is put into nextparam (useful for /c [text])
REM Note that VERIFY>NUL is used to ensure that no errors are returned. SHIFT
REM can result in unpredictable errors which may (or may not) occur later in
REM the routine when referencing non-existent parameters.
REM ECHO got hereE
REM ECHO %0 %*
SET parloop=1
SET param=0
:param_loop
REM IF DEFINED debug ECHO Testing param "%2" for match with "%1"
IF {%2}=={} VERIFY>NUL&GOTO :eof
IF {%2}=={/%1} SET /A param=parloop
REM IF /I {%2}=={-%1} SET /A param=parloop
SHIFT /2
IF NOT %param%==0 (
	SET nextparam=%2
	REM IF DEFINED debug ECHO Found match, nextparam is "!nextparam!"
	VERIFY>NUL&GOTO :eof
)
SET /A parloop=parloop+1
GOTO :param_loop

:openhtml
REM the 'openhtml' word here could be anything...
IF DEFINED debug ECHO Opening %1
START "openhtml" "%~1" 2>NUL
GOTO :EOF

:lower_case
REM converts variable passed to it on command line to lower case, also strips
REM any quotes and any spaces. The returned value is in variable lcase
IF DEFINED debug ECHO %0 %*
SET lcase=%1
SET lcase=%lcase:"=%
SET lcase=%lcase:A=a%
SET lcase=%lcase:B=b%
SET lcase=%lcase:C=c%
SET lcase=%lcase:D=d%
SET lcase=%lcase:E=e%
SET lcase=%lcase:F=f%
SET lcase=%lcase:G=g%
SET lcase=%lcase:H=h%
SET lcase=%lcase:I=i%
SET lcase=%lcase:J=j%
SET lcase=%lcase:K=k%
SET lcase=%lcase:L=l%
SET lcase=%lcase:M=m%
SET lcase=%lcase:N=n%
SET lcase=%lcase:O=o%
SET lcase=%lcase:P=p%
SET lcase=%lcase:Q=q%
SET lcase=%lcase:R=r%
SET lcase=%lcase:S=s%
SET lcase=%lcase:T=t%
SET lcase=%lcase:U=u%
SET lcase=%lcase:V=v%
SET lcase=%lcase:W=w%
SET lcase=%lcase:X=x%
SET lcase=%lcase:Y=y%
SET lcase=%lcase:Z=z%
SET lcase=%lcase: =%
GOTO :eof

:today_mdy
REM by REMS from http://www.petri.co.il/forums/showthread.php?t=19445
REM Should work with all normal short date formats that use numeric values only and do not begin with the year.
REM Will also work with formats that precede the date with dayofweek
ECHO. | date | %findprog% "(mm" > NUL
IF ERRORLEVEL 1 (CALL :Parsedate DD MM) ELSE (CALL :Parsedate MM DD)
VERIFY>NUL
GOTO :EOF
:Parsedate
FOR /F "tokens=1-4 delims=/.- " %%A IN ('date /t') DO IF %%D!==! (
	     SET %1=%%A&SET %2=%%B&SET YYYY=%%C
	) else (
	     SET DOW=%%A&SET %1=%%B&SET %2=%%C&SET YYYY=%%D
	)
)
REM ECHO %MM%-%DD%-%YYYY%
SET mdy=%MM%-%DD%-%YYYY%
GOTO :EOF

:errortext
SET ERRORTEXT=
IF e%1==e1 SET ERRORTEXT=failure at second call of vshadow/vscsc
IF e%1==e2 SET ERRORTEXT=unable to access mapped drive - ensure source drive is formatted with NTFS or add line 'SET vss=n' to configfile
IF e%1==e3 SET ERRORTEXT=Windows Volume Shadow Services not operating correctly - if it has worked on this machine before, try rebooting, otherwise do Start/Run services.msc and check for Volume Shadow Copy which should be started, then retry.
IF e%1==e4 SET ERRORTEXT=cannot find excludelist file on local machine - check SET excludelist= setting
IF e%1==e5 SET ERRORTEXT=missing dependency file
IF e%1==e7 SET ERRORTEXT=vshadow/vscsc -da was unable to delete existing shadow copies - reboot to fix
IF e%1==e8 SET ERRORTEXT=cannot find private key file on local machine - check SET key= setting and/or .ppk file location
IF e%1==e9 SET ERRORTEXT=cannot connect to Server
IF e%1==e10 SET ERRORTEXT=cannot find configuration file
IF e%1==e11 SET ERRORTEXT=needs to be run with administrative privileges when using VSS
IF e%1==e12 SET ERRORTEXT=cannot run under this version of Windows
IF e%1==e13 SET ERRORTEXT=cannot complete
IF e%1==e14 SET ERRORTEXT=cannot connect to rdiff-backup on Server
IF e%1==e15 SET ERRORTEXT=server undefined, or invalid configuration file
IF e%1==e16 SET ERRORTEXT=cannot decode short date format
IF e%1==e17 SET ERRORTEXT=recovering - cannot identify source directory as one used by TimeDicer
IF e%1==e18 SET ERRORTEXT=recovering/listing - an error occurred when recovering, or listing versions of, the specified file
IF e%1==e19 SET ERRORTEXT=no valid sources found, please check configuration file
IF e%1==e20 SET ERRORTEXT=error at second invocation of vshadow/vscsc - try running with /d option
IF e%1==e30 SET ERRORTEXT=error performing rdiff-backup
IF e%1==e50 SET ERRORTEXT=error performing rdiff-backup + error with vshadow/vscsc
IF %1 GTR 100 SET ERRORTEXT=internal error - please report
GOTO :EOF

:help
IF "%2" NEQ "b" (
	REM show the help file in browser
	"%MYAPPDATA%\%thislc%-man.html"
	GOTO :eof
)
